/*
    Randomly display links to a few pages on the site -
    extracted from sitemap.htm

    Optionally excludes paths that occur in the logfile
    for a recent access.

    Copyright 2010-2021, R.Harmsen, all right reserved

    10 May 2021: Removed last vestiges of Adsense code.
        Was already disabled since 2012.

    4 June 2021: Adapted for nginx and Ubuntu Server,
	and for HTML5 instead of 4.

    4 February 2023: Improved parameter handling, after the
    	example of ../dstqview/dstqview.c

    5 February 2023: Better CSS mechanism.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>

#include "../cgitools/cgitools.h"
#include "../fstrfind/fstrfind.h"

enum progmode_enum
{
   PROGMODE_RANDOM = 0,
   PROGMODE_RARELY = 1,
};
struct progdata_s
{
   enum progmode_enum progmode;
   int  loglines;
   char *scriptname;
   /* For an explanation of parameter flashscr see
      https://rudhar.com/sfreview/flashscr/en.htm */
   int  flashscr;
} progdata;

static void invocpar (struct progdata_s *p_progdata);
static void showform (struct progdata_s *p_progdata);
static void htmlhead (struct progdata_s *p_progdata);
static void htmlfoot (struct progdata_s *p_progdata);

static int InLogExtract (char *buf, char *LogExtract);

int main (int argc, char **argv)
{
   progdata.scriptname = "../cgi-bin/rndmview.cgi";

   invocpar(&progdata);
   htmlhead(&progdata);
   showform(&progdata);
   htmlfoot(&progdata);

   return 0;
}

static void invocpar (struct progdata_s *p_progdata)
{
   char *query;
   char **Vars = NULL, **V;

   /* Default values */
   p_progdata->progmode = PROGMODE_RANDOM;
   p_progdata->loglines = 1000;
   p_progdata->flashscr = 0;

   query = getenv("QUERY_STRING");
   if (query)
   {
      Vars = CgiConvVars(query, NULL);

      for (V = Vars; V[0]; V += 2)
      {
         if (strcmp(V[0], "progmode") == 0 && strlen(V[1]))
         {
            if (strcmp("random", V[1]) == 0)
               p_progdata->progmode = PROGMODE_RANDOM;
            else
               p_progdata->progmode = PROGMODE_RARELY;
         }
         else if (strcmp(V[0], "loglines") == 0 && strlen(V[1]))
         {
            p_progdata->loglines = abs(atoi(V[1]));
            if (p_progdata->loglines > 9999)
                p_progdata->loglines = 9999;
         }
         else if (strcmp(V[0], "flashscr") == 0 && strlen(V[1]))
         {
            p_progdata->flashscr = atoi(V[1]);
         }
      }
   }
}

static char HeadText1[] =
   "<!doctype html>\n"
   "<html lang=en>\n"
   "<head>\n"
   "<meta name=viewport content=\"width=device-width, initial-scale=1\">\n"
   /* 20 October 2019: ISO-8859-1 will be UTF-8 from now on,
   because all HTML has been converted */
   "<meta charset=UTF-8>\n"
   ;
static char HeadText2[] =
   "<title>";
static char HeadText3[] =
   "</title>\n"
   "</head>\n"
   "<body>\n"
   "\n<div class=rh-article-noad>\n"
   ;

static void htmlhead (struct progdata_s *p_progdata)
{
   printf("Content-type: text/html\n\n");
   printf("%s", HeadText1);
   if (p_progdata->flashscr)
      printf("%s",
         "<link rel=stylesheet href=\"../cgi-bin/colschem.cgi?url=mainmenu.css\">\n");
   else
      printf("%s",
         "<link rel=stylesheet href=\"../css/mainmenu.css\">\n"
         "<link rel=stylesheet href=\"../cgi-bin/colschem.cgi\">\n");

   printf("%s%s%s\n",
     HeadText2,
     p_progdata->progmode == PROGMODE_RARELY ?
        "Rarely viewed pages from this site" :
        "Randomly selected pages from this site",
     HeadText3);

   if (p_progdata->progmode == PROGMODE_RARELY)
      printf("<h1>Rarely viewed pages from this site</h1>\n");
   else
      printf("<h1>Randomly selected pages from this site</h1>\n");

   fflush(stdout);
}

static void htmlfoot (struct progdata_s *p_progdata)
{
   printf("</div>\n<div class=rh-navifooter-noad>\n");
   printf("<div class=rh-navifooter-content>\n\<hr>\n");
   printf("<p class=pagelink>\n"
      "<a target=_top href=\"../index/viewview.htm\">A view on views</a>\n"
      "</p>\n"
      "<a target=_top href=\"../index.htm\">This site's entry point</a>\n"
      "</p>\n");

   printf("</div>\n");
   printf("<button class=navbutn>Naviga</button>\n");
   printf("</div>\n</body>\n</html>\n");
   fflush(stdout);
}

static char *SitemapFilename = "../html/index/sitemap.htm";

#define DRAND() ((double)random() / (double)0x7fffffffL)
#define SRAND() (srandom((unsigned)time(NULL) + getpid()))

#define NUMENTRIES 16
static void showform (struct progdata_s *p_progdata)
{
   int i, j;
   FILE *fp;
   struct stat st;
   char LogExtract[L_tmpnam];

   LogExtract[0] = '\0';

   if (stat(SitemapFilename, &st))
   {
      printf("<p>Error: cannot stat() file<br>%s</p>\n", SitemapFilename);
      return;
   }

   fp = fopen(SitemapFilename, "r");
   if (!fp)
   {
      printf("<p>Error: cannot fopen() file<br>%s</p>\n", SitemapFilename);
      return;
   }

   /* Safety first */
   LogExtract[0] = '\0';
   if (p_progdata->progmode == PROGMODE_RARELY)
   {
      char cmd[256];
      tmpnam(LogExtract);
      if (LogExtract[0])
      {
         sprintf(cmd,
            "cat /var/log/nginx/access.log.1 /var/log/nginx/access.log |"
            "tail -n %d | fgrep -v -f lstviewd/stoplist | tail -n %d > %s",
            4 * p_progdata->loglines, p_progdata->loglines, LogExtract);
         system(cmd);
      }
   }

   printf("<form action=\"javascript:location.reload();\" method=post>\n");
   printf("<input type=submit name=Refresh value=\"Refresh (F5)\">\n");
   printf("</form>\n\n");

   SRAND();

   for (i = 0, j = 0; i < NUMENTRIES && j < NUMENTRIES * 5; i++, j++)
   {
      long pos1, pos2, len;
      char *buf = NULL, *b = NULL;

      /* Set a random starting point within the sitemap file */
      fseek(fp, (long)((double)st.st_size * DRAND()),SEEK_SET);

      /* Search next <a href, then the end tag </a> */
      pos1 = pos2 = -999;
      pos1 = fstrfindex(fp, "<a href", 1, 1, 1);
      if (pos1 >= 0)
         pos2 = fstrfindex(fp, "</a>", 1, 1, 1);
      if (pos1 >= 0 && pos2 >= 0)
      {
         len = pos2 + 4 - pos1;
         buf = calloc(1, len + 1);
      }
      if (buf)
      {
         fseek(fp, pos1, SEEK_SET);
         fread(buf, len, 1, fp);
         buf[len] = '\0';

         b = strstr(buf, "\">");
         /* Check for empty link texts, these are directories
            that should be skipped. Later on, I also put text
            in links to folders (in the sitemap.htm), so the test
            needs to be done in a different way.
          */
         if ((b && b + 2 == buf + len - 4)
            ||
            strstr(buf, "<i>Folder:</i>")
	         ||
	         /* Filter Deventer moordzaak */
	         ((strstr(buf, "politics/devmrdzk")) &&
	            DRAND() > 0.20)
	         ||
	         /* Filter Demmink, Micha Kat etc. */
	         ((strstr(buf, "politics/intnlgov")) &&
	            DRAND() > 0.70)
	         ||
	         strstr(buf, "/index/index-")
            ||
            (p_progdata->progmode == PROGMODE_RARELY && InLogExtract(buf, LogExtract))
         )
         {
	          i--;
         }
         else
         {
            b = strstr(buf, "=\"");
            if (b)
            {
               printf("<p><a target=_top href%s</p>\n", b);
            }
         }

         if (buf)
            free(buf);
      }
   }

   fclose(fp);
   {
      /* Example of longest form action line:
         <form action="../cgi-bin/rndmview.cgi?progmode=rarely&amp;loglines=9999&amp;flashscr=1" method=post>\n
         length 101 bytes. So we take 128 to be safe. */
      char buf[128];
      char *b = buf;

      b += sprintf(b, "<form action=\"%s?progmode=%s", p_progdata->scriptname,
         p_progdata->progmode == PROGMODE_RANDOM ? "random" : "rarely");

      if (p_progdata->progmode == PROGMODE_RARELY)
         b += sprintf(b, "&amp;loglines=%d", p_progdata->loglines);

      if (p_progdata->flashscr)
         b += sprintf(b, "&amp;flashscr=1");

      b += sprintf(b, "\" method=post>\n");

      printf("%s", buf);
   }
   printf(
      "<input type=submit name=Randomise value=\"New selection\"> \n"
      "</form>\n\n");

   if (p_progdata->progmode == PROGMODE_RARELY && LogExtract[0])
   {
      unlink(LogExtract);
   }
}

/* Check if the html file, found randomly from sitemap.htm,
   occurs in recent entries of the Apache of nginx log file.
   If so, it is not selected for the optional functionality
   "show rarely viewed pages".
 */
static int InLogExtract (char *buf, char *LogExtract)
{
   char *URL = NULL;
   char *b = NULL, *e = NULL;
   pid_t pid;
   int status = 0;
   int retval = 0;

   URL = calloc(1, strlen(buf) + 1);
   if (!URL)
      return 0;

   b = strstr(buf, "\"../");
   if (b)
      e = strstr(b + 1, "\"");

   if (!b || !e || b > e)
      return 0;

   sprintf(URL, "GET %-*.*s ", (int)(e - b - 3), (int)(e - b - 3), b + 3);

   pid = fork();

   if (pid < 0)
   {
      printf("<p>Cannot fork!<p>\n");
   }
   else if (pid == 0) // Child process
   {
      int ret;

      /* Redirect standard output and standard error to /dev/null,
         to supress any output from grep (option -q and -s
         don't seem to work) */
      close(1);
      close(2);
      open("/dev/null", O_RDWR);
      open("/dev/null", O_RDWR);

      if ((ret = execlp("grep", "-q", "-s", "-m1", URL, LogExtract, NULL)))
      {
         printf("<p>Exec returned %d, errno = %d, %s, URL = %s</p>\n", ret, errno, strerror(errno), URL);
         exit(0);
      }
   }
   else // Parent process
   {
      wait(&status);
#ifdef DEBUG
      printf("<p>Status = %d, exitstatus = %d</p>\n", status, WEXITSTATUS(status));
#endif
      if (WEXITSTATUS(status) == 0)
         retval = 1;
   }

#ifdef DEBUG
   if (retval && URL)
   {
      printf("<p>Skipped %s</p>\n", URL);
   }
#endif

   if (URL)
      free(URL);

   return retval;
}

